#include <cstdlib>
#include <msp/fs/dir.h>
#include <msp/strings/format.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "androidarchiver.h"
#include "androidassetpackagingtool.h"
#include "androidccompiler.h"
#include "androidcxxcompiler.h"
#include "androidlinker.h"
#include "androidmanifestgenerator.h"
#include "androidtools.h"
#include "apkbuilder.h"
#include "architecture.h"
#include "builder.h"
#include "jarsigner.h"

using namespace std;
using namespace Msp;

// TODO Mark problems somewhere instead of throwing exceptions

unsigned parse_version(const std::string &version_str)
{
	vector<string> version_parts = split(version_str, '.');
	unsigned version = lexical_cast<unsigned>(version_parts[0])<<16;
	if(version_parts.size()>1)
		version += lexical_cast<unsigned>(version_parts[1])<<8;
	if(version_parts.size()>2)
		version += lexical_cast<unsigned>(version_parts[2]);
	return version;
}


AndroidDevKit::AndroidDevKit(Builder &b, const string &type):
	builder(b)
{
	string var = format("ANDROID_%s_ROOT", type);
	if(const char *value = getenv(var.c_str()))
		root = value;
	else
		throw runtime_error(var+" must be set");

	FS::Path platforms_dir = root/"platforms";
	builder.get_logger().log("files", format("Traversing %s", platforms_dir.str()));
	list<string> platforms = list_filtered(platforms_dir, "^android-[1-9][0-9]*$");

	for(list<string>::const_iterator i=platforms.begin(); i!=platforms.end(); ++i)
		supported_api_levels.insert(lexical_cast<unsigned>(i->substr(8)));
}

void AndroidDevKit::select_api_level(unsigned api)
{
	if(!supported_api_levels.count(api))
		throw invalid_argument("AndroidDevKit::select_api_level");

	init_api_level(api);
}


AndroidSdk::AndroidSdk(Builder &b):
	AndroidDevKit(b, "SDK")
{
	find_build_tools_dir();
}

void AndroidSdk::find_build_tools_dir()
{
	FS::Path bt_dir = root/"build-tools";
	builder.get_logger().log("files", format("Traversing %s", bt_dir.str()));
	list<string> tool_versions = list_files(bt_dir);

	string use_tools;
	unsigned latest_version = 0;
	for(list<string>::const_iterator i=tool_versions.begin(); i!=tool_versions.end(); ++i)
	{
		unsigned version = parse_version(*i);
		if(version>latest_version)
		{
			use_tools = *i;
			latest_version = version;
		}
	}

	if(use_tools.empty())
		throw runtime_error("No build tools found");

	build_tools_dir = bt_dir/use_tools;
	builder.get_logger().log("tools", format("Android build tools are in %s", build_tools_dir.str()));
}

void AndroidSdk::init_api_level(unsigned api)
{
	if(!supported_api_levels.count(api))
		throw invalid_argument("AndroidSdk::select_api_level");

	platform_jar = root/"platforms"/format("android-%d", api)/"android.jar";
}


AndroidNdk::AndroidNdk(Builder &b, const Architecture &a):
	AndroidDevKit(b, "NDK"),
	architecture(a)
{
	find_toolchain_dir();
}

void AndroidNdk::find_toolchain_dir()
{
	FS::Path toolchains_dir = root/"toolchains";
	builder.get_logger().log("files", format("Traversing %s", toolchains_dir.str()));
	string prefix = architecture.get_cross_prefix()+"-";
	list<string> toolchains = list_filtered(toolchains_dir, "^"+prefix);

	string use_toolchain;
	unsigned latest_version = 0;
	for(list<string>::const_iterator i=toolchains.begin(); i!=toolchains.end(); ++i)
	{
		string version_str = i->substr(prefix.size());
		string compiler = "gcc";
		if(!isdigit(version_str[0]))
		{
			unsigned j;
			for(j=1; (j<version_str.size() && !isdigit(version_str[j])); ++j) ;
			compiler = version_str.substr(0, j);
			version_str = version_str.substr(j);
		}

		if(compiler!="gcc")
			continue;

		unsigned version = parse_version(version_str);
		if(version>latest_version)
		{
			use_toolchain = *i;
			latest_version = version;
		}
	}

	if(use_toolchain.empty())
		throw runtime_error("No applicable toolchains found");

	const Architecture &native_arch = builder.get_native_arch();

	FS::Path tc_archs_dir = toolchains_dir/use_toolchain/"prebuilt";
	builder.get_logger().log("files", format("Traversing %s", tc_archs_dir.str()));
	string use_arch = native_arch.best_match(list_files(tc_archs_dir));

	if(use_arch.empty())
		throw runtime_error("No matching toolchain found");

	bin_dir = toolchains_dir/use_toolchain/"prebuilt"/use_arch/"bin";
	builder.get_logger().log("tools", format("Android toolchain binaries are in %s", bin_dir.str()));
}

void AndroidNdk::init_api_level(unsigned api)
{
	if(!supported_api_levels.count(api))
		throw invalid_argument("AndroidSdk::select_api_level");

	FS::Path platform_archs_dir = root/"platforms"/format("android-%d", api);
	builder.get_logger().log("files", format("Traversing %s", platform_archs_dir.str()));
	list<string> platform_archs = list_filtered(platform_archs_dir, "^arch-");
	for(list<string>::iterator i=platform_archs.begin(); i!=platform_archs.end(); ++i)
		i->erase(0, 5);
	string use_arch = architecture.best_match(platform_archs);

	if(use_arch.empty())
		throw runtime_error("No matching platform found");

	platform_sysroot = platform_archs_dir/("arch-"+use_arch);
	builder.get_logger().log("tools", format("Android NDK sysroot is %s", platform_sysroot.str()));
}


AndroidTools::AndroidTools(Builder &builder, const Architecture &arch):
	sdk(builder),
	ndk(builder, arch)
{
	const set<unsigned> &sdk_api_levels = sdk.get_supported_api_levels();
	const set<unsigned> &ndk_api_levels = ndk.get_supported_api_levels();
	unsigned highest_common = 0;
	for(set<unsigned>::const_reverse_iterator i=sdk_api_levels.rbegin(); (!highest_common && i!=sdk_api_levels.rend()); ++i)
		if(ndk_api_levels.count(*i))
			highest_common = *i;

	if(!highest_common)
		throw runtime_error("No usable Android platforms found");

	sdk.select_api_level(highest_common);
	ndk.select_api_level(highest_common);

	add_tool(new AndroidCCompiler(builder, arch, ndk));
	add_tool(new AndroidCxxCompiler(builder, arch, ndk));
	add_tool(new AndroidLinker(builder, arch, ndk));
	add_tool(new AndroidArchiver(builder, arch, ndk));
	add_tool(new AndroidManifestGenerator(builder));
	add_tool(new AndroidAssetPackagingTool(builder, sdk));
	add_tool(new ApkBuilder(builder));
	add_tool(new JarSigner(builder));
}
